import unittest

from django.test.runner import DebugSQLTextTestResult, DiscoverRunner

from pdbr.__main__ import RichPdb, post_mortem


class PDBRDebugResult(unittest.TextTestResult):
    _pdbr = RichPdb()

    def addError(self, test, err):
        super().addError(test, err)
        self._print(test, err)

    def addFailure(self, test, err):
        super().addFailure(test, err)
        self._print(test, err)

    def _print(self, test, err):
        self.buffer = False
        self._pdbr.message(f"\n{test}")
        self._pdbr.error("%s: %s", err[0].__name__, err[1])
        post_mortem(err[2])


class PdbrDiscoverRunner(DiscoverRunner):
    def get_resultclass(self):
        if self.debug_sql:
            return DebugSQLTextTestResult
        return PDBRDebugResult



import sys

from pdbr.__main__ import RichPdb


def run_ipython_shell():
    try:
        from IPython.terminal.interactiveshell import TerminalInteractiveShell
        from IPython.terminal.ipapp import TerminalIPythonApp
        from prompt_toolkit.history import FileHistory
        from traitlets import Type

        TerminalInteractiveShell.simple_prompt = False
    except ModuleNotFoundError as error:
        raise type(error)(
            "In order to use pdbr shell, install IPython with pdbr[ipython]"
        ) from error

    class PdbrTerminalInteractiveShell(TerminalInteractiveShell):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            if RichPdb._ipython_history_file:
                self.debugger_history = FileHistory(RichPdb._ipython_history_file)

        @property
        def debugger_cls(self):
            return RichPdb

    class PdbrTerminalIPythonApp(TerminalIPythonApp):
        interactive_shell_class = Type(
            klass=object,  # use default_value otherwise which only allow subclasses.
            default_value=PdbrTerminalInteractiveShell,
            help=(
                "Class to use to instantiate the TerminalInteractiveShell object. "
                "Useful for custom Frontends"
            ),
        ).tag(config=True)

    app = PdbrTerminalIPythonApp.instance()
    app.initialize()
    sys.exit(app.start())



import sys
from telnetlib import Telnet

from rich.file_proxy import FileProxy

from pdbr.helpers import run_ipython_shell


def shell():
    import getopt

    _, args = getopt.getopt(sys.argv[1:], "mhc:", ["command="])

    if not args:
        run_ipython_shell()
    else:
        from pdbr.__main__ import main

        main()


def telnet():
    from pdbr.__main__ import RichPdb

    pdb_cls = RichPdb()
    if len(sys.argv) < 3:
        pdb_cls.error("Usage : pdbr_telnet hostname port")
        sys.exit()

    class MyTelnet(Telnet):
        def fill_rawq(self):
            """
            exactly the same with Telnet.fill_rawq,
            buffer size is just changed from 50 to 1024.
            """
            if self.irawq >= len(self.rawq):
                self.rawq = b""
                self.irawq = 0
            buf = self.sock.recv(1024)
            self.msg("recv %r", buf)
            self.eof = not buf
            self.rawq = self.rawq + buf

    console = pdb_cls.console
    sys.stdout = FileProxy(console, sys.stdout)
    sys.stderr = FileProxy(console, sys.stderr)
    try:
        host = sys.argv[1]
        port = int(sys.argv[2])
        with MyTelnet(host, port) as tn:
            tn.interact()
    except BaseException as e:
        pdb_cls.error(e)
        sys.exit()



import atexit
import configparser
from pathlib import Path

from pdbr._pdbr import rich_pdb_klass

try:
    import readline
except ImportError:
    try:
        from pyreadline3 import Readline

        readline = Readline()
    except ModuleNotFoundError:
        readline = None
except AttributeError:
    readline = None


def set_history_file(history_file):
    """
    This is just for Pdb,
    For Ipython, look at RichPdb.pt_init
    """
    if readline is None:
        return
    try:
        readline.read_history_file(history_file)
        readline.set_history_length(1000)
    except FileNotFoundError:
        pass
    except OSError:
        pass

    atexit.register(readline.write_history_file, history_file)


def set_traceback(theme):
    from rich.traceback import install

    install(theme=theme)


def read_config():
    style = None
    theme = None
    store_history = ".pdbr_history"

    config = configparser.ConfigParser()
    config.sections()

    setup_filename = "setup.cfg"
    global_config_path = Path.home() / ".config" / "pdbr" / setup_filename
    cwd_config_path = Path.cwd() / setup_filename
    config_path = cwd_config_path.exists() and cwd_config_path or global_config_path

    config.read(config_path)
    if "pdbr" in config:
        if "style" in config["pdbr"]:
            style = config["pdbr"]["style"]

        if "theme" in config["pdbr"]:
            theme = config["pdbr"]["theme"]

        if "use_traceback" in config["pdbr"]:
            if config["pdbr"]["use_traceback"].lower() == "true":
                set_traceback(theme)
        else:
            set_traceback(theme)

        if "store_history" in config["pdbr"]:
            store_history = config["pdbr"]["store_history"]

    history_file = str(Path.home() / store_history)
    set_history_file(history_file)
    ipython_history_file = f"{history_file}_ipython"

    return style, theme, history_file, ipython_history_file


def debugger_cls(
    klass=None, console=None, context=None, is_celery=False, show_layouts=True
):
    if klass is None:
        try:
            from IPython.terminal.debugger import TerminalPdb

            klass = TerminalPdb
        except ImportError:
            from pdb import Pdb

            klass = Pdb

    RichPdb = rich_pdb_klass(
        klass,
        console=console,
        context=context,
        is_celery=is_celery,
        show_layouts=show_layouts,
    )
    style, theme, history_file, ipython_history_file = read_config()
    RichPdb._style = style
    RichPdb._theme = theme
    RichPdb._history_file = history_file
    RichPdb._ipython_history_file = ipython_history_file

    return RichPdb


def _pdbr_cls(console=None, context=None, return_instance=True, show_layouts=True):
    klass = debugger_cls(console=console, context=context, show_layouts=show_layouts)
    if return_instance:
        return klass()
    return klass


def _rdbr_cls(return_instance=True):
    try:
        from celery.contrib import rdb

        rdb.BANNER = """\
{self.ident}: Type `pdbr_telnet {self.host} {self.port}` to connect

{self.ident}: Waiting for client...
"""
    except ModuleNotFoundError as error:
        raise type(error)("In order to install celery, use pdbr[celery]") from error

    klass = debugger_cls(klass=rdb.Rdb, is_celery=True, show_layouts=False)
    if return_instance:
        return klass()
    return klass



from pdbr.__main__ import RichPdb, celery_set_trace, pm, post_mortem, run, set_trace
from pdbr._cm import apdbr_context, pdbr_context

__all__ = [
    "set_trace",
    "run",
    "pm",
    "post_mortem",
    "celery_set_trace",
    "RichPdb",
    "pdbr_context",
    "apdbr_context",
]



from starlette.middleware.errors import ServerErrorMiddleware

from pdbr._cm import apdbr_context


class PdbrMiddleware(ServerErrorMiddleware):
    async def __call__(self, scope, receive, send) -> None:
        async with apdbr_context(suppress_exc=False, debug=self.debug):
            await super().__call__(scope, receive, send)



import os
import pdb
import sys

from .utils import _pdbr_cls, _rdbr_cls

os.environ["PYTHONBREAKPOINT"] = "pdbr.set_trace"

RichPdb = _pdbr_cls(return_instance=False, show_layouts=False)


def set_trace(*, console=None, header=None, context=None, show_layouts=False):
    pdb_cls = _pdbr_cls(console=console, context=context, show_layouts=show_layouts)
    if header is not None:
        pdb_cls.message(header)
    pdb_cls.set_trace(sys._getframe().f_back)


def run(statement, globals=None, locals=None):
    RichPdb().run(statement, globals, locals)


def post_mortem(traceback=None, value=None):
    _, sys_value, sys_traceback = sys.exc_info()
    value = value or sys_value
    traceback = traceback or sys_traceback

    if traceback is None:
        raise ValueError(
            "A valid traceback must be passed if no exception is being handled"
        )

    p = RichPdb()
    p.reset()
    if value:
        p.error(value)
    p.interaction(None, traceback)


def pm():
    post_mortem(sys.last_traceback)


def celery_set_trace(frame=None):
    pdb_cls = _rdbr_cls()
    if frame is None:
        frame = sys._getframe().f_back
    return pdb_cls.set_trace(frame)


def main():
    pdb.Pdb = RichPdb
    pdb.main()


if __name__ == "__main__":
    main()



from rich.containers import Lines
from rich.errors import NotRenderableError
from rich.layout import Layout
from rich.panel import Panel


class ConsoleLayoutMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class ConsoleLayout(metaclass=ConsoleLayoutMeta):
    def __init__(self, console):
        self.console = console
        self.layout = self._prep_layout()

    def _prep_layout(self):
        layout = Layout()
        right_body = Layout(name="right_body", ratio=1)

        layout.split(
            Layout(name="left_body", ratio=2),
            right_body,
            splitter="row",
        )

        right_body.split(
            Layout(name="up_footer", ratio=2), Layout(name="bottom_footer", ratio=1)
        )
        return layout

    def print(self, message, code, stack_trace, vars, **kwargs):
        try:
            self.layout["left_body"].update(code)
            self.layout["up_footer"].update(Panel(vars, title="Locals"))

            self.layout["bottom_footer"].update(
                Panel(Lines(stack_trace), title="Stack", style="white on blue")
            )

            self.console.print(self.layout, **kwargs)
            self.console.print(message, **kwargs)
        except NotRenderableError:
            self.console.print(message, **kwargs)
